/*********************************************************************

         FILE:  basic_demod.c

   PROGRAMMER:  Roger Hammons

                Hughes Network Systems
                11717 Exploration Lane
                Germantown, MD 20878
                e-mail:  rhammons@hns.com
                tel:     (301) 428-2784
                fax:     (301) 428-2750
                
      PURPOSE:  This file contains the Basic_Demod module.
                The intent of this module is provide a simple model 
                for an UTRA compatible demodulator that can be used 
                by all participating ETSI members to evaluate turbo 
                code performance over realistic channels but under 
                identical assumptions with regard to demodulation
                algorithms.  These basic demodulation algorithms 
                were agreed upon by all participants in the turbo 
                code ad hoc committee in the October 1998 meeting 
                in Stockholm, Sweden.  

  DESCRIPTION:  The following describe the major assumptions and 
                features of the Basic_Demod module.  The user is
                also referred to Tdoc SMG2 UMTS-L1 493/98 for 
                larger simulation issues such as signal scaling, 
                noise scaling, power control operations, etc. 

                (1) The basic demod module processes one slot 
                    at a time.  The input parameter symbols_per_slot 
                    specifies the number of QPSK symbols per slot, 
                    including pilot, data, and DTX symbols.  The 
                    agreed upon slot structure was 4 contiguous 
                    pilot symbols at start of the slot, followed 
                    by a variable number of contiguous data symbols, 
                    and terminated by a variable number of all-zero 
                    DTX symbols.  The number of data symbols per 
                    slot depends on the channel interleaver buffer 
                    size and can vary by one QPSK symbol frome slot 
                    to slot within the same interleaver buffer.  
                (2) The pilot symbols are all (1/sqrt(2))(1+i), 
                    the "00" constellation point for the off-axis 
                    QPSK constellation.  The DTX symbols have
                    constant value equal to complex 0.  The 
                    Basic_Demod module checks for DTX symbol by 
                    looking at real and imaginary components of 
                    the input data_signal.  If both components 
                    have magnitudes less than a zero tolerance 
                    value (defined in the header file 
                    basic_demod_types.h), a DTX symbol is assumed 
                    to be present.  Such symbols are not counted 
                    in signal power calculations.   
                (3) i.i.d. AWGN noise samples are provided for 
                    all symbols, including pilot, data, and DTX. 
                    AWGN samples must be supplied for all channel  
                    taps -- not just for fingers only.  This is a 
                    change from the Stockholm agreement necessitated 
                    by deficiency with respect to the agreed upon 
                    algorithms first noted by Mitsubishi.  (See 
                    revision history item dated 10/29/98.)  
                (4) In the modulation process, it may be necessary 
                    to add an extra data bit to make a full number 
                    of QPSK symbols.  The outside simulation is 
                    responsible for this bookkeeping, and the 
                    existence of such an extra information bit 
                    is transparent to the Basic_Demod module.  
                    Contrary to suggestion made in Stockholm, 
                    it should not affect average signal power 
                    or Eb/No calculations.  The extra information
                    bit should however be ignored in bit error rate 
                    determination.  
                (5) The channel model is assumed to provide 7 taps 
                    of channel response.  If the channel model 
                    provides less than 7 taps, the remaining taps 
                    should be initialized to complex 0.  The basic 
                    demod will select the best num_fingers of these 
                    as the processed fingers (best meaning highest 
                    power).  
                (6) Channel estimation is over consecutive slots.
                    This requires idling one slot before data 
                    becomes valid; hence, the first interleaver 
                    buffer must be thrown away.  The power control 
                    loop also requires idle time to settle.  Based 
                    on these considerations, it was agreed that 
                    simulations should ignore the first 8 frames 
                    of data.  Note that Basic_Demod processes data 
                    only on a slot-by-slot basis so cannot and 
                    does not perform the frame trashing function.  
                (7) The output power control command pc_flag must 
                    be used by the outside simulation to scale 
                    the input modulated data.  
                (8) Average signal power per slot is based on non-
                    DTX data after fading but before AWGN noise is 
                    added.  Average noise power is computed over 
                    all symbols on all fingers.  These estimates 
                    are to be used in the turbo code MAP constituent 
                    decoders. 
                (9) Functions not handled by Basic_Demod that must 
                    be provided by the controlling simulation 
                    include:  turbo code encoding, QPSK modulation, 
                    channel interleaving, slot framing, turbo 
                    decoding including SNR estimation, power control 
                    loop, bit error rate/block error rate statistics,   
                    and simulation-long averaging to estimate Eb/No 
                    for performance curve plotting.                             
                
   DISCLAIMER:  While a good faith effort has been made to implement 
                the agreed upon algorithms accurately, it is expected 
                that the participating ETSI members will test and 
                verify to their own satisfaction the correctness of 
                the implementation. 
                 

 REVISION HISTORY:  
   10-19-98   rh   Initial stub to specify interface.
   10-20-98   rh   Add ave_noise_pwr to interface.
   10-22-98   rh   Completed C code for basic demod algorithms.  
   10-29-98   rh   Revised channel estimation to correct defficiency 
                   in agreed upon algorithms as pointed out by 
                   Mitsubishi in their e-mail dated 10-28-98.  
                   Change is to evaluate channel estimates for all 
                   taps instead of just fingers.  Otherwise, there 
                   is a potential misalignment due to different 
                   fingers being selected on adjacent slots. This 
                   requires an interface change:  Noise samples 
                   must be supplied for all channel taps.  (Before 
                   it was allowed to supply noise samples for 
                   fingers only.) 
  10-30-98    me   Converted from dynamic finger selection to fixed
                   finger selection according to suggestions made
                   by NTT DoCoMo

  11-24-98    me   Placed the pilot bits at the end of the slot, increased
                   the span of the channel estimator filter to include the
                   present and the three previous slots, increased the 
                   pilot overhead by a factor of four. 

  12-09-98    me   Re-placed the pilot bits at the beginning of the slot,
                   but implemented buffering of slots to improve the
                   reliability of channel estimation.

*********************************************************************/

#include <math.h>
//#include "basic_demod_types.h"

static struct DATA old1_tap_chnl_est[MAXCHNLTAPS];
static struct DATA old2_tap_chnl_est[MAXCHNLTAPS];
static struct DATA old3_tap_chnl_est[MAXCHNLTAPS];
static struct DATA old4_tap_chnl_est[MAXCHNLTAPS]; 
static struct DATA old5_tap_chnl_est[MAXCHNLTAPS];
static struct DATA old6_tap_chnl_est[MAXCHNLTAPS];
static struct DATA old1_noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS]; 
static struct DATA old2_noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS]; 
static struct DATA old3_noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS];
static struct DATA old4_noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS];
static struct DATA old5_noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS];
static struct DATA old6_noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS]; 
static double old1_ave_signal_pwr, old2_ave_signal_pwr, old3_ave_signal_pwr,
             old4_ave_signal_pwr;
static double old1_ave_noise_pwr, old2_ave_noise_pwr, old3_ave_noise_pwr,
             old4_ave_noise_pwr;

void Basic_Demod( struct DATA data_signal[],
                  const int    *symbols_per_slot,
                  const struct NOISE noise_signal[],
                  const int    *num_fingers,
                  const struct CHANNEL chnl_response[],
                  struct DATA soft_output[],
                  int    *pc_flag,
                  double  *ave_signal_pwr,
                  double  *ave_noise_pwr  )
{
   /******************************************************************
      Local variable declarations
   ******************************************************************/

   int fig; 
   int findex[MAXCHNLTAPS];
   int nonDTX[MAXSYMPERSLOT];
   int nonDTX_count;
   int sym; 
   int tap; 

   double ave_tap_pwr[MAXCHNLTAPS];
   double inv_root_two;
   double pilot_factor;
   double pilot_pwr;
   double pilot_tap_pwr;
   double overhead_increase=4.0;

   struct DATA conjugate_pilot;
   struct DATA faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS];
   struct DATA noise;
   struct DATA noisy_faded_tap[MAXSYMPERSLOT][MAXCHNLTAPS];
   struct DATA pathgain;
   struct DATA product;
   struct DATA reference;
   struct DATA old1, old2, old3, old4, old5, old6;
   struct DATA tap_chnl_est[MAXCHNLTAPS];
   struct DATA tap_chnl_est_ave[MAXCHNLTAPS];

   /******************************************************************
      Function declarations
   ******************************************************************/

   double Csqmag( struct DATA *complexnum );

   void  Cmult(  struct DATA *left, 
                 struct DATA *right, 
                 struct DATA *product );
   void  Czero(  struct DATA *complexnum );
   void  Cadd(   struct DATA *left, 
                 struct DATA *right, 
                 struct DATA *sum );
   void  Ccopy(  struct DATA *complexnum,
                 struct DATA *copiednum );
   void  Cscale( double factor, 
                 struct DATA *complexnum );
   void  Cscalarmult( double factor, 
                 struct DATA *complexnum,
                 struct DATA *result );
   void  Cconj(  struct DATA *complexnum, 
                 struct DATA *conjugate );

   /******************************************************************
      Compute useful constants. 
   ******************************************************************/

   pilot_factor = 1. / (double)( PILOT_END - PILOT_START + 1 );

   inv_root_two = (double)(1./sqrt(2.));
   conjugate_pilot.Re =  inv_root_two;
   conjugate_pilot.Im = -inv_root_two;



   /******************************************************************
    Increase the pilot overhead by a factor of 4. Pilots should be
    at the beginning of the slot
    ******************************************************************/

   for (sym=PILOT_START; sym<= PILOT_END; sym++)
     {
       data_signal[sym].Re=(double)sqrt(overhead_increase)*data_signal[sym].Re;
       data_signal[sym].Im=(double)sqrt(overhead_increase)*data_signal[sym].Im;
     }

   /******************************************************************
      Apply fading across all taps of all symbols
   ******************************************************************/

   for( sym = 0; sym < *symbols_per_slot; ++sym ) {
      pathgain.Re = chnl_response[sym].Re1;
      pathgain.Im = chnl_response[sym].Im1;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][0]) );

      pathgain.Re = chnl_response[sym].Re2;
      pathgain.Im = chnl_response[sym].Im2;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][1]) );

      pathgain.Re = chnl_response[sym].Re3;
      pathgain.Im = chnl_response[sym].Im3;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][2]) );

      pathgain.Re = chnl_response[sym].Re4;
      pathgain.Im = chnl_response[sym].Im4;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][3]) );

      pathgain.Re = chnl_response[sym].Re5;
      pathgain.Im = chnl_response[sym].Im5;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][4]) );

      pathgain.Re = chnl_response[sym].Re6;
      pathgain.Im = chnl_response[sym].Im6;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][5]) );

      pathgain.Re = chnl_response[sym].Re7;
      pathgain.Im = chnl_response[sym].Im7;
      Cmult( &pathgain, &(data_signal[sym]), &(faded_tap[sym][6]) );
   }

   /******************************************************************
      Compute average power of faded tap data for each tap,
      excising DTX values.  
   ******************************************************************/

   nonDTX_count = 0;
   for( sym = PILOT_END+1; sym < *symbols_per_slot; ++sym ) {
      if( (fabs(data_signal[sym].Re) > ZEROTOL) ||  
          (fabs(data_signal[sym].Im) > ZEROTOL) ) {
         nonDTX[sym] = 1;
         nonDTX_count++;
      }
      else { 
         nonDTX[sym] = 0;
      }
   }

   for( tap = 0; tap < MAXCHNLTAPS; ++tap ) {
      ave_tap_pwr[tap] = 0.;
      for( sym = PILOT_END+1; sym < *symbols_per_slot; ++sym ) {
         if( nonDTX[sym] ) {
            ave_tap_pwr[tap] += Csqmag( &(faded_tap[sym][tap]) );
         }
      }
      ave_tap_pwr[tap] = ave_tap_pwr[tap] / nonDTX_count;
   }
  
   /******************************************************************
      Compute total average power summed over all taps.
      (Output variable)
   ******************************************************************/

   *ave_signal_pwr = 0.;
   for( tap = 0; tap < MAXCHNLTAPS; ++tap ) {
      *ave_signal_pwr += ave_tap_pwr[tap];
   }

   /******************************************************************
      Sort the taps based on multipath channel attenuation 
   ******************************************************************/

   /* Fixed finger assignment: Choose the best average fingers throughout
      the simulation   */

   findex[0]=0;
   findex[1]=1;
   findex[2]=2;
   findex[3]=4;
   findex[4]=3;
   findex[5]=5;
   findex[6]=6;

   /******************************************************************
      Perform power control calculation on faded finger data.
      (Output variable)
   ******************************************************************/

   pilot_pwr = 0.;
   for( fig = 0; fig < *num_fingers; ++fig ) {
      pilot_tap_pwr = 0.;
      for( sym = PILOT_START; sym <= PILOT_END; ++sym ) {
         pilot_tap_pwr += Csqmag( &(faded_tap[sym][findex[fig]]) );
      }
      pilot_pwr += pilot_tap_pwr;
   }
   pilot_pwr = pilot_factor * pilot_pwr;

   if( pilot_pwr >= overhead_increase ) *pc_flag = POWER_DN;
   else                   *pc_flag = POWER_UP;

   /******************************************************************
      Add noise to faded tap data (all symbols, all taps).
      Compute average noise power over all taps. (Output variable)
   ******************************************************************/

   *ave_noise_pwr = 0;
   tap = 0; 
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re1;
         noise.Im = noise_signal[sym].Im1;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }
   tap = 1;
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re2;
         noise.Im = noise_signal[sym].Im2;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }
   tap = 2;
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re3;
         noise.Im = noise_signal[sym].Im3;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }
   tap = 3;
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re4;
         noise.Im = noise_signal[sym].Im4;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }
   tap = 4;
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re5;
         noise.Im = noise_signal[sym].Im5;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }
   tap = 5;
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re6;
         noise.Im = noise_signal[sym].Im6;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }
   tap = 6;
   if( tap < MAXCHNLTAPS ) {
      for( sym = 0; sym < *symbols_per_slot; ++sym ) {
         noise.Re = noise_signal[sym].Re7;
         noise.Im = noise_signal[sym].Im7;
         Cadd( &(faded_tap[sym][tap]),
               &noise,  
               &(noisy_faded_tap[sym][tap]) );
         *ave_noise_pwr += Csqmag( &noise );
      }
   }

   *ave_noise_pwr = *ave_noise_pwr
                    /(double)( MAXCHNLTAPS * *symbols_per_slot );

   /*****************************************************************
      Buffer the noise and signal power estimates.
      ***************************************************************/

   old4_ave_signal_pwr=old3_ave_signal_pwr;
   old3_ave_signal_pwr=old2_ave_signal_pwr;
   old2_ave_signal_pwr=old1_ave_signal_pwr;
   old1_ave_signal_pwr=(*ave_signal_pwr);

   old4_ave_noise_pwr=old3_ave_noise_pwr;
   old3_ave_noise_pwr=old2_ave_noise_pwr;
   old2_ave_noise_pwr=old1_ave_noise_pwr;
   old1_ave_noise_pwr=(*ave_noise_pwr);


   /******************************************************************
      Compute tap channel estimates from noisy faded tap 
      data for current slot. 
   ******************************************************************/

   for( tap = 0; tap < MAXCHNLTAPS; ++tap ) {
      Czero( &(tap_chnl_est[tap]) );
      for( sym = PILOT_START; sym <= PILOT_END; ++sym ) {
         Cmult( &(noisy_faded_tap[sym][tap]), &conjugate_pilot, &product );
         Cadd(  &(tap_chnl_est[tap]), &product, &(tap_chnl_est[tap]) );
      }
      Cscale( pilot_factor, &(tap_chnl_est[tap]) );
   }
      
   /*****************************************************************
   Prepare the buffered signal and noise power estimates as current output 
   ********************************************************************/

      (*ave_signal_pwr)=old4_ave_signal_pwr;
      (*ave_noise_pwr)=old4_ave_noise_pwr;

   /*******************************************************************
     Buffer the noisy faded data
     *****************************************************************/

     for (sym=0; sym<MAXSYMPERSLOT; sym++)
       for (tap=0; tap<MAXCHNLTAPS; tap++)
       {
 Ccopy( &(old5_noisy_faded_tap[sym][tap]), &(old6_noisy_faded_tap[sym][tap]));
 Ccopy( &(old4_noisy_faded_tap[sym][tap]), &(old5_noisy_faded_tap[sym][tap]));
 Ccopy( &(old3_noisy_faded_tap[sym][tap]), &(old4_noisy_faded_tap[sym][tap]));
 Ccopy( &(old2_noisy_faded_tap[sym][tap]), &(old3_noisy_faded_tap[sym][tap]));
 Ccopy( &(old1_noisy_faded_tap[sym][tap]), &(old2_noisy_faded_tap[sym][tap]));
 Ccopy( &(noisy_faded_tap[sym][tap]), &(old1_noisy_faded_tap[sym][tap]) );
       }
   
   /******************************************************************
      Filter tap channel estimates over current, three future and
      two previous slots, filter coefficients are 0.3 0.8 1 1 0.8 0.3.
   ******************************************************************/
    

     for( tap = 0; tap < MAXCHNLTAPS; ++tap ) 
      {
	Ccopy( &(old5_tap_chnl_est[tap]), &(old6_tap_chnl_est[tap]) ); 
	Ccopy( &(old4_tap_chnl_est[tap]), &(old5_tap_chnl_est[tap]) ); 
	Ccopy( &(old3_tap_chnl_est[tap]), &(old4_tap_chnl_est[tap]) );
	Ccopy( &(old2_tap_chnl_est[tap]), &(old3_tap_chnl_est[tap]) );
	Ccopy( &(old1_tap_chnl_est[tap]), &(old2_tap_chnl_est[tap]) );
	Ccopy( &(tap_chnl_est[tap]), &(old1_tap_chnl_est[tap]) );
      }

     for( tap = 0; tap < MAXCHNLTAPS; ++tap )
      {
	Cscalarmult(0.3, &(old1_tap_chnl_est[tap]), &(old1));
	Cscalarmult(0.8, &(old2_tap_chnl_est[tap]), &(old2));
        Cscalarmult(1.0, &(old3_tap_chnl_est[tap]), &(old3));
        Cscalarmult(1.0, &(old4_tap_chnl_est[tap]), &(old4));
        Cscalarmult(0.8, &(old5_tap_chnl_est[tap]), &(old5));
        Cscalarmult(0.3, &(old6_tap_chnl_est[tap]), &(old6));
	Cadd( &(old1), &(old2), 
            &(tap_chnl_est_ave[tap]) );
        Cadd( &(tap_chnl_est_ave[tap]), 
            &(old3), 
            &(tap_chnl_est_ave[tap]) );
        Cadd( &(tap_chnl_est_ave[tap]), 
            &(old4), 
            &(tap_chnl_est_ave[tap]) );
	Cadd( &(tap_chnl_est_ave[tap]), 
            &(old5), 
            &(tap_chnl_est_ave[tap]) );
        Cadd( &(tap_chnl_est_ave[tap]), 
            &(old6), 
            &(tap_chnl_est_ave[tap]) );
        Cscale( 1.0/4.2, &(tap_chnl_est_ave[tap]) );
     } 
   
      
   /******************************************************************
      Perform maximal ratio combining of finger data using 
      averaged finger channel estimates.
      (Output variable)
   ******************************************************************/

   for( sym = 0; sym < *symbols_per_slot; ++sym ) {
      Czero( &(soft_output[sym]) );
      for( fig = 0; fig < *num_fingers; ++fig ) {
         Cconj( &(tap_chnl_est_ave[findex[fig]]), &reference );
         Cscale( 1.0/sqrt(overhead_increase), &reference );
         Cmult( &(old4_noisy_faded_tap[sym][findex[fig]]),
                &reference, 
                &product );
         Cadd( &(soft_output[sym]), &product, &(soft_output[sym]) );
      }
   }
}


/*********************************************************************
                     AUXILIARY SUBROUTINES
**********************************************************************
                        Ccopy 
*********************************************************************/

 void Ccopy(  struct DATA *complexnum, struct DATA *copiednum )
{
   copiednum->Re = complexnum->Re;
   copiednum->Im = complexnum->Im;
}


/*********************************************************************
                        Czero 
*********************************************************************/

 void Czero( struct DATA *complexnum )
{
   complexnum->Re = 0.;
   complexnum->Im = 0.;
}


/*********************************************************************
                        Csqmag 
*********************************************************************/

 double Csqmag( struct DATA *complexnum )
{
   double value;
   value = complexnum->Re*complexnum->Re 
           + complexnum->Im*complexnum->Im;
   return( value );
}


/*********************************************************************
                        Cadd 
*********************************************************************/

 void  Cadd(   struct DATA *left, 
              struct DATA *right, 
              struct DATA *sum )
{
   sum->Re = left->Re + right->Re;
   sum->Im = left->Im + right->Im;
}


/*********************************************************************
                        Cmult 
*********************************************************************/

 void  Cmult(  struct DATA *left, 
              struct DATA *right, 
              struct DATA *product ) 
{
   struct DATA temp;
   temp.Re = left->Re * right->Re - left->Im * right->Im;
   temp.Im = left->Re * right->Im + left->Im * right->Re;
   Ccopy( &temp, product );
}


/*********************************************************************
                        Cscale 
*********************************************************************/

 void  Cscale( double factor, struct DATA *complexnum )
{
   complexnum->Re = factor * complexnum->Re;
   complexnum->Im = factor * complexnum->Im;
}

/*********************************************************************
                        Cscalarmult
*********************************************************************/

 void  Cscalarmult( double factor, struct DATA *complexnum, struct DATA *result )
{
   result->Re = factor * complexnum->Re;
   result->Im = factor * complexnum->Im;
}


/*********************************************************************
                        Cconj 
*********************************************************************/

 void  Cconj(  struct DATA *complexnum, struct DATA *conjugate )
{
   conjugate->Re =  complexnum->Re;
   conjugate->Im = -complexnum->Im;
}


